/*
 *  Copyright 2018 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

#ifndef MODULES_DESKTOP_CAPTURE_APP_CAPTURER_X11_H_
#define MODULES_DESKTOP_CAPTURE_APP_CAPTURER_X11_H_

#include <X11/X.h>
#include <X11/Xlib.h>
#include <memory>
#include <string>

#include "modules/desktop_capture/app_capturer.h"
#include "modules/desktop_capture/shared_desktop_frame.h"
#include "modules/desktop_capture/desktop_capture_options.h"
#include "modules/desktop_capture/desktop_capturer.h"
#include "modules/desktop_capture/desktop_geometry.h"
#include "modules/desktop_capture/x11/shared_x_display.h"
#include "modules/desktop_capture/x11/x_server_pixel_buffer.h"
#include "rtc_base/constructormagic.h"
#include "rtc_base/scoped_ref_ptr.h"

namespace webrtc {

class ScreenCapturerProxy : DesktopCapturer::Callback {
public:
  ScreenCapturerProxy()
    : screen_capturer_(DesktopCapturer::CreateScreenCapturer(DesktopCaptureOptions::CreateDefault())) {
    screen_capturer_->SelectSource(kFullDesktopScreenId);
    screen_capturer_->Start(this);
  }
  void CaptureFrame() { screen_capturer_->CaptureFrame(); }
  std::unique_ptr<DesktopFrame> GetFrame() { return std::move(frame_); }

   // Callback interface
  virtual void OnCaptureResult(DesktopCapturer::Result result,
                               std::unique_ptr<DesktopFrame> frame) {
    frame_ = std::move(frame);
  }

protected:
  std::unique_ptr<DesktopCapturer> screen_capturer_;
  std::unique_ptr<DesktopFrame> frame_;
};

class AppCapturerX11 : public AppCapturer {
public:
  AppCapturerX11(const DesktopCaptureOptions& options);
  virtual ~AppCapturerX11();

  static AppCapturer* Create(const DesktopCaptureOptions& options);
  static std::unique_ptr<DesktopCapturer> CreateRawAppCapturer(const DesktopCaptureOptions& options);

  // AppCapturer interface.
  virtual bool GetAppList(AppList* apps) override;
  virtual bool SelectApp(ProcessId processId) override;
  virtual bool BringAppToFront() override;

  // DesktopCapturer interface.
  virtual void Start(Callback* callback) override;
  virtual void CaptureFrame() override;
  virtual bool SelectSource(SourceId id) override
  {
    return SelectApp(static_cast<ProcessId>(id));
  }

protected:
  Display* GetDisplay() { return x_display_->display(); }
  bool UpdateRegions();

  void FillDesktopFrameRegionWithColor(DesktopFrame* pDesktopFrame,Region rgn, uint32_t color);
private:
  Callback* callback_;
  ProcessId selected_process_;

  // Sample Mode
  ScreenCapturerProxy screen_capturer_proxy_;
  // Mask of foreground (non-app windows in front of selected)
  Region rgn_mask_;
  // Region of selected windows
  Region rgn_visual_;
  // Mask of background (desktop, non-app windows behind selected)
  Region rgn_background_;

  rtc::scoped_refptr<SharedXDisplay> x_display_;
  RTC_DISALLOW_COPY_AND_ASSIGN(AppCapturerX11);
};

}  // namespace webrtc

#endif  // MODULES_DESKTOP_CAPTURE_APP_CAPTURER_X11_H_
